// 用来维护线程
package com.kob.matchingsystem.service.impl.utils;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.util.LinkedMultiValueMap;
import org.springframework.util.MultiValueMap;
import org.springframework.web.client.RestTemplate;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.locks.ReentrantLock;

@Component
public class MatchingPool extends Thread {
    // 把所有用户存下来
    private static List<Player> playerList = new ArrayList<>();
    // 定义一个锁, Reentrant Lock 是可重入锁
    private final ReentrantLock lock = new ReentrantLock();
    // 定义微服务传值 URL
    private static final String startGameUrl = "http://127.0.0.1:3001/pk/start/";
    // 定义 RestTemplate 用来微服务数据通信
    private static RestTemplate restTemplate;

    @Autowired
    private void setRestTemplate(RestTemplate restTemplate) {
        MatchingPool.restTemplate = restTemplate;
    }

    public void addPlayer(Integer userId, Integer rating, Integer botId) {
        lock.lock();
        try {
            // 一开始匹配等待时间是 0
            playerList.add(new Player(userId, rating, botId, 0));
        } finally {
            lock.unlock();
        }
    }

    public void removePlayer(Integer userId) {
        lock.lock();
        try {
            // 先创建一个新列表,将没有删的用户先存下来
            List<Player> newPlayerList = new ArrayList<>();
            // 枚举所有的 player 并存入 newPlayerList
            for (Player player : playerList) {
                // 如果当前 player 的 id 不等于要删除的用户 id,则表示此用户保留,将要保留的用户信息存入新列表
                // 否则不存入新列表
                if (!player.getUserId().equals(userId))
                    newPlayerList.add(player);
            }
            // 将删除用户后的新 newPlayerList 赋值给 playerList
            playerList = newPlayerList;
        } finally {
            lock.unlock();
        }
    }

    // 将所有玩家的等待时间 +1
    private void increaseWaitingTime() {
        for (Player player : playerList) {
            player.setWaitingTime(player.getWaitingTime() + 1);
        }
    }

    // 判断两名玩家是否匹配
    private boolean checkMatched(Player a, Player b) {
        int ratingDelta = Math.abs(a.getRating() - b.getRating()); // 两名玩家的天梯积分之差
        // a b 两名玩家的等待时间最小值
        // min 是两方任意一方接受等待时间就可以匹配, max 是两方都接受的匹配
        int minWaitingTime = Math.min(a.getWaitingTime(), b.getWaitingTime());
        // 匹配结果可被 a 接受: a b 的分差小于等于 a和b 最小等待时间乘以 10
        // 匹配结果可被 b 接受: a b 的分差小于等于 a和b 最小等待时间乘以 10
        return ratingDelta <= minWaitingTime * 10;
    }

    // 返回 a 和 b 的匹配结果
    private void sendResult(Player a, Player b) {
        MultiValueMap<String, String> data = new LinkedMultiValueMap<>();
        // 返回两名玩家的 userId 和 botId
        data.add("a_id", a.getUserId().toString());
        data.add("a_bot_id", a.getBotId().toString());
        data.add("b_id", b.getUserId().toString());
        data.add("b_bot_id", b.getBotId().toString());
        restTemplate.postForObject(startGameUrl, data, String.class);
    }

    // 尝试匹配所有玩家
    private void matchPlayers() {
        // TODO 后端调试
        // System.out.println("match players: " + playerList.toString());
        // havaMatched 表示玩家已经匹配过了
        boolean[] haveMatched = new boolean[playerList.size()];
        for (int i = 0; i < playerList.size(); i++) {
            if (haveMatched[i])
                continue; // 如果当前枚举到的玩家已经匹配过了,则跳过改玩家
            for (int j = i + 1; j < playerList.size(); j++) { // j 从 i+1 开始枚举
                if (haveMatched[j])
                    continue;
                // 如果 i 和 j 都没有匹配,则将 a 和 b 玩家取出来
                Player a = playerList.get(i);
                Player b = playerList.get(j);
                // 判断 a 和 b 能否匹配:如果匹配,则将结果返回,并将 a 和 b 的位置置为 true
                if (checkMatched(a, b)) {
                    haveMatched[i] = haveMatched[j] = true; // 置为已匹配
                    sendResult(a, b); // 返回匹配玩家结果
                    break;
                }
            }
        }

        // 匹配成功后,需要将匹配过的玩家从匹配池移除
        List<Player> newPlayerList = new ArrayList<>();
        for (int i = 0; i < playerList.size(); i++) {
            // 如果当前枚举到的玩家没有匹配,则将该玩家存入新的 list
            if (!haveMatched[i])
                newPlayerList.add(playerList.get(i));
        }
        // 将新的 newPlayerList 重新赋值 playerList
        playerList = newPlayerList;
    }

    // 实现线程
    @Override
    public void run() {
        while (true) {
            try {
                Thread.sleep(1000);
                lock.lock();
                try {
                    // 每隔一秒钟调用“等待时间增加”函数
                    increaseWaitingTime();
                    // 尝试匹配玩家
                    matchPlayers();
                } finally {
                    lock.unlock();
                }
            } catch (InterruptedException e) {
                // 如果捕获到异常,则输出异常并终止运行
                e.printStackTrace();
                break;
            }
        }
    }
}
